from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
import time, re, sys, os, shutil, argparse, json, urllib.parse
import pickle
from datetime import datetime
from .downloader import *
from .utils import *

root = os.path.abspath(os.path.dirname(__file__))+"/"
THEMES=["auto"]+os.listdir(root+"themes/")

def main(args,bar=False):
    if bar:
        bar_output=sys.stdout
    else:
        bar_output=open(os.devnull,"w")

    if not args["theme"] in THEMES:
        raise_error('Invalid theme name `'+args["theme"]+'`')

    if args["axel"]:
        delay=0.1
    else:
        delay=1

    nd_klass = get_downloader(args["url"])
    if nd_klass:
        nd = nd_klass(args["url"],delay=delay,bar_output=bar_output)
    else:
        raise_error("URL is not supported")

    nd.extract_info()

    if args["episode"]:
        if not re.match(r'^\d*$',args["episode"]) or int(args["episode"])>nd.info["num_parts"] or int(args["episode"])<0:
            raise_error("Incorrect episode number `"+args["episode"]+"`")
        nd.mark_all("skip")
        nd.mark_part("unskip",int(args["episode"]))
        args["episode"]=int(args["episode"])

    try:
        args["name"] = args["name"].format("",ncode=nd.ncode,title=re.sub(r'[\\|/|:|?|.|"|<|>|\|]', '', nd.info["title"]))
    except KeyError:
        raise_error("Incorrect format template")
    now=datetime.now()
    args["name"] = now.strftime(args["name"])
    db_data = {}

    if nd.info["num_parts"] == 0 or args["episode"]:
        if args["dir"]:
            ndir=os.path.abspath(args["dir"])+"/"
        else:
            ndir=os.getcwd()+"/"
    else:
        if args["dir"]:
            ndir=os.path.abspath(args["dir"])+"/"+args["name"]+"/"
        else:
            ndir=os.getcwd()+"/"+args["name"]+"/"
        if os.path.isfile(ndir+"static/db.json") and not args["renew"]:
            with open(ndir+"static/db.json","r") as f:
                db_data = json.load(f)
            nd.mark_all("skip")
            if nd.info["num_parts"] > db_data["num_parts"]:
                nd.mark_part("unskip",db_data["num_parts"])
            for i in nd.info["epis"].keys():
                if not str(i) in db_data["epis"]:
                    nd.mark_part("unskip",i)
                elif nd.info["epis"][i]["time"] > datetime.fromisoformat(db_data["epis"][str(i)]):
                    nd.mark_part("unskip",i)

    try:
        nd.extract_novels()
    except NovelDLException as e:
        if e.return_id() == 1:
            e.console_message()
        else:
            raise e

	# Load themes
    if args["theme"]=="auto":
        args["theme"] = nd.auto_theme
    THEME_DIR=root+"themes/"+args["theme"]+"/"
    conf_file=os.path.join(THEME_DIR,"config.json")
    conf = {}
    if os.path.isfile(conf_file):
        with open(conf_file,"r") as f:
            conf = json.load(f)
    MEDIAS=[""]
    if conf.get("medias"):
        MEDIAS=conf["medias"]
    if not args["media"] in MEDIAS:
        MEDIAS.remove("")
        raise_error("Invalid media type\nAvailable medias in this theme: ({})".format(" ".join(MEDIAS)))
    if args["media"]:
        htmls={"base":"base_{}.html".format(args["media"]),"index":"index_{}.html".format(args["media"]),"single":"single_{}.html".format(args["media"])}
    else:
        htmls={"base":"base.html","index":"index.html","single":"single.html"}

    if conf.get('parent'):
        env_paths=[THEME_DIR,os.path.join(root,"themes",conf['parent'])]
        lstatic = os.listdir(THEME_DIR+"static/")
        static_files = [THEME_DIR+"static/"+i for i in lstatic]+[os.path.join(env_paths[1],"static",i) for i in os.listdir(os.path.join(env_paths[1],"static")) if not i in lstatic]
    else:
        env_paths=THEME_DIR
        static_files = [os.path.join(THEME_DIR,"static",i) for i in os.listdir(os.path.join(THEME_DIR,"static"))]
    env=Environment(loader=FileSystemLoader(env_paths,encoding='utf8'))
    try:
        htmls={i:env.get_template(htmls[i]) for i in htmls}
    except TemplateNotFound as e:
        raise_error("Cannot load theme file: "+e.name)

    # Create directory
    if not os.path.isdir(ndir):
        os.makedirs(ndir)

    if nd.info["type"] == "short" or args["episode"]:
        style={}
        script={}
        for file in static_files:
            base,ext=os.path.splitext(file)
            if ext==".css":
                with open(file,"r",encoding="utf-8") as f:
                    style[os.path.basename(base)]=f.read()
            if ext==".js":
                with open(file,"r",encoding="utf-8") as f:
                    script[os.path.basename(base)]=f.read()

        if args["episode"]:
            contents=htmls['single'].render(title=nd.novels[args["episode"]][0],author=nd.info["author"],contents=nd.novels[args["episode"]][1],style=style,script=script,lines=len(nd.novels[args["episode"]][1]),url=args["url"])
        else:
            contents=htmls['single'].render(title=nd.info["title"],author=nd.info["author"],contents=nd.novels[0][1],style=style,script=script,lines=len(nd.novels[0][1]),url=args["url"])
        with open(ndir+args["name"]+".html", "w") as f:
            f.write(contents)
    else:
        # Gen index.html
        contents=htmls['index'].render(title=nd.info["title"],author=nd.info["author"],desc=nd.info["desc"],index=nd.info["index"],total=nd.info["num_parts"],url=nd.indexurl)
        with open(ndir+"index.html","w",encoding="utf-8") as f:
            f.write(contents)

		# Copy statics
        if not os.path.islink(ndir+"static"):
            if os.path.isdir(ndir+"static"):
                shutil.rmtree(ndir+"static")
            elif os.path.isfile(ndir+"static"):
                os.remove(ndir+"static")
            #os.symlink(THEME_DIR+"static",ndir+"static")
            os.mkdir(ndir+"static")
            for file in static_files:
                shutil.copyfile(file,ndir+"static/"+os.path.basename(file))


        with open(ndir+"static/db.json","w") as f:
            json.dump(nd.gen_db(db_data),f, ensure_ascii=False, indent=4, sort_keys=True, separators=(',', ': '))

        for part in nd.novels:
            contents=htmls['base'].render(title=nd.info["title"],author=nd.info["author"],subtitle=nd.novels[part][0],part=part,total=nd.info["num_parts"],contents=nd.novels[part][1],lines=len(nd.novels[part][1]),index=nd.info["index"],epis=nd.info["epis"],url=nd.info["epis"][part]["url"])
            with open(ndir+str(part)+".html", "w", encoding="utf-8") as f:
                f.write(contents)

def command_line():
    parser=argparse.ArgumentParser()
    parser.add_argument('url',help="URL")
    parser.add_argument('-d',"--dir",default="",help="set output directory")
    parser.add_argument('-n',"--name",default="{title}",help="set output directory/file name")
    parser.add_argument('-r',"--renew",action='store_true',help="force to update all files")
    parser.add_argument('-a',"--axel",action='store_true',help="turn on axceleration mode")
    parser.add_argument('-e',"--episode",default="",help="set download single episode as short novel")
    parser.add_argument('-t',"--theme",default="auto",help="set novel's theme")
    parser.add_argument('-m',"--media",default="",help="generate html supporting only one media type")
    parser.add_argument('-q',"--quiet",action='store_false',help="suppress non-messages")
    args=parser.parse_args()
    args=args.__dict__
    output=args.pop("quiet")
    try:
        main(args,bar=output)
    except NovelDLException as e:
        e.console_message()
        sys.exit(-1)
    else:
        if output:
            print("Successfully downloaded")

def args(url="",save_dir="",renew=False,axel=False,episode="",theme="auto",media=""):
    return {"url": url,"dir": save_dir,"renew": renew,"axel": axel,"episode": episode,"theme": theme,"media": media}

if __name__=="__main__":
    command_line()



import os
import json
import sys
import argparse
from .utils import get_config_path, raise_error
from .info import (
    __appname__,
    __description__,
)

root = os.path.abspath(os.path.dirname(__file__))+"/"

class ConfigManager(object):
    def __init__(self):
        self.config=dict()
        self.load_default()
        self.config_dir = get_config_path()
        self.config_file = os.path.join(self.config_dir,"settings.json")

    def load_default(self):
        self.config = {
            "default_theme": "auto",
            "theme_path": [os.path.join(root,"themes")],
            "default_delay": 1,
            "min_delay": 0.1,
            "retries": 3,
            "user_agant": "",
            "output_path": ".",
            "output_format": "{title}",
            "symlink_static": False
        }

    def load_usercfg(self):
        if not os.path.isdir(self.config_dir):
            return
        try:
            with open(os.path.join(self.config_dir,self.config_file),"r") as f:
                conf=json.load(f)
            self.config.update(self.verify_config(conf))
        except json.decoder.JSONDecodeError as e:
            raise_error("Config load error: "+e.msg)

    def save_usercfg(self):
        if not os.path.isdir(self.config_dir):
            os.mkdir(self.config_dir)
            os.mkdir(os.path.join(self.config_dir,"themes"))
            self.config["theme_path"]+=[os.path.join(self.config_dir,"themes")]
        with open(self.config_file,"w") as f:
            json.dump(self.config, f, ensure_ascii=False, indent=4)

    def generate_theme_option(self,args=sys.argv):
        parser=argparse.ArgumentParser(prog=__appname__.lower(),description=__description__)

    def verify_config(self,sd):

        for key in sd:
            if key not in self.config:
                sd.pop(key)
            elif type(sd[key]) != type(self.config[key]):
                sd.pop(key)

        theme_paths=self.config['theme_path']
        for path in sd.get('theme_path'):
            if not os.path.isdir(path):
                sd['theme_path'].remove(path)
            theme_paths.append(path)

        valid_themes=["auto"]
        [valid_themes.append(j) for i in theme_paths for j in os.listdir(i) if not j.startswith('.') and not j in valid_themes]

        rules_dict = {
            "default_theme": valid_themes,
        }

        for key, valid_values in list(rules_dict.items()):
            if sd[key] not in valid_values:
                sd.pop(key)

        dd=self.config['default_delay']
        if sd.get('default_delay'):
            if sd['default_delay']<0:
                sd.pop('default_delay')
            else:
                dd=sd['default_delay']
        if sd.get('min_delay'):
            if sd['min_delay']<0 or sd['min_delay'] > dd:
                sd.pop('min_delay')

        if sd.get('retries') and sd.get('retries') < 0:
            sd.pop('min_delay')

        if sd.get('output_format'):
            try:
                sd['output_format'].format("",ncode="",title="")
            except:
                sd.pop('output_format')
        return sd

class ArgumentManager(object):
    def __init__(self):
        pass


from .default import NarouND, KakuyomuND

def get_downloader(url):
    klass = {name:func for name,func in globals().items() if name.endswith('ND')}
    for i in klass:
        if klass[i].match_url(url):
            return klass[i]
    else:
        return None



import time, re, sys, json, os
import urllib.parse
from datetime import datetime as dtime
from pytz import timezone
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from bs4 import BeautifulSoup as bs4
from tqdm import tqdm
from ..utils import *


class NovelDownloader(object):
    HEADER = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:61.0) Gecko/20100101 Firefox/61.0"}
    COOKIE = {}

    def __init__(self, url, delay=1, params=None,bar_output=sys.stdout):
        self.status = ["INIT"]
        self.bar_output=bar_output
        self._markers = ["dl", "skip"]
        self.url = url
        self.delay = delay
        self.params = params
        self.session = requests.Session()
        self.set_headers(self.HEADER)
        self.set_cookies(self.COOKIE)
        retries = Retry(total=3, backoff_factor=1)
        self.session.mount("https://", HTTPAdapter(max_retries=retries))
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.info = {"title": "", "desc": "", "author": [], "type": "", "num_parts": 0, "index": [], "epis": {}}
        self._mark = []
        self.novels = {}

    def __del__(self):
        self.close()

    def initialize(self):
        self.info = {"title": "", "desc": "", "author": "", "type": "", "num_parts": 0, "index": [], "epis": {}}

    def close(self):
        self.session.close()

    def set_cookies(self, cookies):
        self.session.cookies.update(cookies)

    def get_cookies(self):
        return self.session.cookies

    def set_headers(self, headers):
        self.session.headers.update(headers)

    def get_headers(self):
        return self.session.headers

    def get(self, url, params=None):
        if not params:
            params = self.params
        result = self.session.get(url, params=params)
        return result

    def mark_part(self, com, part):
        if not "INFO" in self.status:
            self.extract_info()
        if part>0 and part<=self.info["num_parts"]:
            if com == "skip" and part in self._mark:
                self._mark.remove(part)
            elif com == "unskip" and not part in self._mark:
                self._mark.append(part)

    def mark_all(self,com):
        if not "INFO" in self.status:
            self.extract_info()
        if com == "skip":
            self._mark.clear()
        elif com == "unskip":
            self._mark = list(range(1,self.info["num_parts"]+1))

    def match_url(url):
        pass

    def extract_info(self):
        try:
            self.initialize()
            self._real_extract_info()
        except KeyboardInterrupt:
            raise_error("Operation canceled by user")
        except requests.exceptions.ConnectionError as e:
            raise_error("Network Error")
        self.status.append("INFO")
        self._mark = list(range(1, self.info["num_parts"]+1))

    def extract_novels(self):
        if not "INFO" in self.status:
            result = self.extract_info()
            if result == -1:
                return -1
        try:
            self._real_extract_novels()
        except KeyboardInterrupt:
            raise_error("Operation was canceled by user",id=1)
        except requests.exceptions.ConnectionError as e:
            raise_error("Network Error",id=1)
        self.status.append("NOVELS")

    def _real_extract_info(self):
        pass

    def _real_extract_novels(self):
        pass

    def gen_db(self,db_data={}):
        db = {"url": self.indexurl, "title": self.info["title"], "num_parts": self.info["num_parts"], "author": self.info["author"], "epis": {}}
        if db_data:
            db["epis"]={int(i):j for i,j in db_data["epis"].items()}
        db["epis"].update({i:self.info["epis"][i]["time"].isoformat() for i in self.novels})
        return db


class NarouND(NovelDownloader):
    COOKIE = {'over18': 'yes'}

    def __init__(self, url, delay=1, params=None, bar_output=sys.stdout):
        self.auto_theme = "narou"
        super().__init__(url, delay, params,bar_output=bar_output)
        ret = urllib.parse.urlparse(url)
        self.ncode = re.match(r'/(n[0-9a-zA-Z]+)', ret.path).group(1)
        self.baseurl = "https://{}".format(ret.hostname)
        self.indexurl = self.baseurl+"/"+self.ncode

    def match_url(url):
        ret = urllib.parse.urlparse(url)
        if ret.hostname == "ncode.syosetu.com" or ret.hostname == "novel18.syosetu.com":
            return True
        else:
            return False

    def _real_extract_info(self):
        data = self.get(self.indexurl)
        top_data = bs4(data.content, "html.parser")
        if top_data.select_one(".maintenance-container"):
            raise_error("Narou is under maintenance")
        self.info["title"] = top_data.select_one("title").text
        author_data = top_data.select_one(".novel_writername")
        if author_data.a:
            self.info["author"] = [author_data.a.text,author_data.a.attrs["href"]]
        else:
            self.info["author"] = [author_data.text[4:][:-1],""]
        index_raw = top_data.select_one(".index_box")
        if index_raw:
            self.info["num_parts"] = len(index_raw.select(".novel_sublist2"))
            self.info["type"] = "serial"
        else:
            self.info["num_parts"] = 0
            self.info["type"] = "short"
            body = top_data.select_one("#novel_honbun")
            l = [bs4(str(i), "html.parser") for i in body("p")]
            [i.p.unwrap() for i in l]
            body = [str(i) for i in l]
            self.novels.update({0:(self.info["title"], body)})
            self.status.append("NOVELS")
            return

        eles = bs4(str(index_raw).replace("\n", ""),"html.parser").contents[0].contents
        c = ""
        cid = 1
        part = 1
        for ele in eles:
            if re.match(r'.+chapter_title', str(ele)):
                self.info["index"].append({"type": "chapter", "id":cid, "text": ele.text})
                c = ele.text
                cid = cid+1
            elif re.match(r'.+novel_sublist2', str(ele)):
                timestamp = dtime.strptime(ele.dt.text.replace("（改）", ""), "%Y/%m/%d %H:%M")
                self.info["index"].append({"type": "episode", "part": part, "text": ele.a.text,"time": timestamp})
                self.info["epis"][part]={"subtitle": ele.a.text, "url": self.baseurl+ele.a.attrs['href'], "chap": c, "time": timestamp}
                part = part+1
        self.info["desc"] = "".join([str(i) for i in top_data.select_one("#novel_ex").contents])

    def _real_extract_novels(self):
        if self.info["num_parts"] == 0:
            return
        with tqdm(total=len(self._mark),file=self.bar_output,unit="parts") as pbar:
            pbar.set_description("Downloading ")
            for part in self._mark:
                res = self.get(self.info["epis"][part]["url"])
                soup = bs4(res.content, "html.parser")
                subtitle = soup.select_one(".novel_subtitle").text
                body = soup.select_one("#novel_honbun")

                l = [bs4(str(i), "html.parser") for i in body("p")]
                [i.p.unwrap() for i in l]
                body = [str(i) for i in l]

                self.novels.update({part:(subtitle, body)})
                pbar.update()
                time.sleep(self.delay)


class KakuyomuND(NovelDownloader):
    def __init__(self,url,delay=1,params=None, bar_output=sys.stdout):
        self.auto_theme = "kakuyomu"
        super().__init__(url,delay,params,bar_output=bar_output)
        ret = urllib.parse.urlparse(url)
        self.ncode = re.match(r'/works/([0-9]+)',ret.path).group(1)
        self.baseurl = "https://{}".format(ret.hostname)
        self.indexurl = self.baseurl+"/works/"+self.ncode

    def match_url(url):
        ret = urllib.parse.urlparse(url)
        if ret.hostname == "kakuyomu.jp":
            return True
        else:
            return False

    def _real_extract_info(self):
        data = self.get(self.indexurl)
        top_data = bs4(data.content, "html.parser")
        index_raw=top_data.select_one(".widget-toc-items")
        raws=index_raw.select("li.widget-toc-episode")
        self.info["num_parts"] = len(raws)
        self.info["type"] = "serial"
        author_data = top_data.select_one("#workAuthor-activityName")
        self.info["author"] = [author_data.a.text,self.baseurl+author_data.a.attrs["href"]]
        self.info["title"] = top_data.select_one("#workTitle").text
        eles = bs4(str(index_raw).replace("\n",""),"html.parser").contents[0].contents
        c = ""
        cid = 1
        part = 1
        for ele in eles:
            if re.match(r'.+widget-toc-chapter',str(ele)):
                self.info["index"].append({"type": "chapter", "id": cid, "text": ele.text})
                c = ele.text
            elif re.match(r'.+widget-toc-episode',str(ele)):
                timestamp=dtime.strptime(ele.a.time.get('datetime'),"%Y-%m-%dT%H:%M:%SZ").astimezone(timezone('Asia/Tokyo'))
                self.info["index"].append({"type": "episode", "part": part, "text": ele.span.text, "time": timestamp})
                self.info["epis"][part]={"subtitle": ele.span.text, "url": self.baseurl+ele.a.attrs["href"], "chap": c, "time": timestamp}
                part=part+1
        desc=top_data.select_one("#introduction")
        if desc.select_one(".ui-truncateTextButton-expandButton"):
            desc.select_one(".ui-truncateTextButton-expandButton").decompose()
            desc.span.unwrap()
        self.info["desc"]="".join([str(i) for i in desc.contents])

    def _real_extract_novels(self):
        with tqdm(total=len(self._mark),file=self.bar_output,unit="parts") as pbar:
            pbar.set_description("Downloading ")
            for part in self._mark:
                res = self.get(self.info["epis"][part]["url"])
                soup = bs4(res.content, "html.parser")
                subtitle=soup.select_one(".widget-episodeTitle").text
                body=soup.select_one(".widget-episodeBody")

                l=[bs4(str(i),"html.parser") for i in body("p")]
                [i.p.unwrap() for i in l]
                body=[str(i) for i in l]

                self.novels.update({part:(subtitle, body)})
                pbar.update()
                time.sleep(self.delay)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages
import os

root=os.path.abspath(os.path.dirname(__file__))

PACKAGE_NAME="novel_dl"
info_path=os.path.join(root,PACKAGE_NAME,"info.py")
ns=dict()
with open(info_path,"r") as f:
	eval(compile(f.read(), info_path, 'exec'),dict(),ns)
__version__=ns['__version__']

with open(os.path.join(root,"README.md"),"r") as f:
	LONG_DESCRIPTION = f.read()

require_path=os.path.join(root,"requirements.txt")
with open(require_path,"r") as f:
	REQUIRES=f.read().splitlines()

setup(
	name=PACKAGE_NAME,
	version=__version__,

	description=ns['__description__'],
	long_description=LONG_DESCRIPTION,
	long_description_content_type="text/markdown",
	url="https://github.com/kazuto28/{pkg}".format(pkg=ns['__appname__'].lower()),
	author=ns['__author__'],
	author_email=ns['__contact__'],
	license=ns['__license__'],

	packages=[PACKAGE_NAME],
	include_package_data=True,
	zip_safe=False,
	install_requires=REQUIRES,
	entry_points="""
		[console_scripts]
		{app}={pkg}.main:command_line
	""".format(app=ns['__appname__'].lower(),pkg=PACKAGE_NAME),

	classifiers=[
		"Environment :: Console",
		"License :: OSI Approved :: MIT License",
		"Natural Language :: Japanese",
		"Natural Language :: English",
		"Operating System :: Unix",
		"Programming Language :: Python :: 3 :: Only",
	],
)



import sys,os
import importlib.util
from .info import __appname__

class NovelDLException(Exception):
    def console_message(self):
        print(self.return_message(),file=sys.stderr)
    def return_message(self):
        return "novel-dl: "+self.args[0]
    def return_id(self):
        return self.args[1]


def raise_error(msg,id=0):
    raise NovelDLException(msg,id)

def get_config_path():
    if os.name == 'nt':
        path = os.getenv('APPDATA')
    else:
        path = os.path.expanduser('~/.config')

    return os.path.join(path, __appname__.lower())

def import_from_file(name,path):
    if not os.path.isfile(path):
        raise FileNotFoundError
    spec=importlib.util.spec_from_file_location(name,path)
    module=importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module
